@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -56,7 +56,7 @@ static void usage(int status)
 Command syntax:\n\
   cluster (info|format|shutdown)\n\
   node (info|list)\n\
-  vdi (list|tree|graph|delete|object)\n\
+  vdi (list|tree|graph|delete|object|setattr|getattr)\n\
 \n\
 Common parameters:\n\
   -a, --address           specify the daemon address (default: localhost)\n\
@@ -165,7 +165,10 @@ struct cluster_cmd_data {
 
 struct vdi_cmd_data {
 	unsigned int index;
-	int snapshot;
+	int snapshot_id;
+	char snapshot_tag[SD_MAX_VDI_TAG_LEN];
+	int exclusive;
+	int delete;
 } vdi_cmd_data = { ~0, };
 
 static int cluster_format(int argc, char **argv)
@@ -241,7 +244,8 @@ static int shutdown_sheepdog(void)
 	return 0;
 }
 
-typedef void (*vdi_parser_func_t)(uint32_t vid, char *name, uint32_t tag, uint32_t flags,
+typedef void (*vdi_parser_func_t)(uint32_t vid, char *name, char *tag,
+				  uint32_t snapid, uint32_t flags,
 				  struct sheepdog_inode *i, void *data);
 
 static int parse_vdi(vdi_parser_func_t func, size_t size, void *data)
@@ -340,7 +344,7 @@ static int parse_vdi(vdi_parser_func_t func, size_t size, void *data)
 			}
 		}
 
-		func(i.vdi_id, i.name, i.snap_id, 0, &i, data);
+		func(i.vdi_id, i.name, i.tag, i.snap_id, 0, &i, data);
 	next:
 		close(fd);
 	}
@@ -348,7 +352,7 @@ static int parse_vdi(vdi_parser_func_t func, size_t size, void *data)
 	return 0;
 }
 
-static void print_vdi_list(uint32_t vid, char *name, uint32_t tag,
+static void print_vdi_list(uint32_t vid, char *name, char *tag, uint32_t snapid,
 			   uint32_t flags, struct sheepdog_inode *i, void *data)
 {
 	int idx;
@@ -381,12 +385,12 @@ static void print_vdi_list(uint32_t vid, char *name, uint32_t tag,
 
 	if (!data || strcmp(name, data) == 0) {
 		printf("%c %-8s %5d %7s %7s %7s %s  %7" PRIx32 "\n",
-		       is_current(i) ? ' ' : 's', name, tag,
+		       is_current(i) ? ' ' : 's', name, snapid,
 		       vdi_size_str, my_objs_str, cow_objs_str, dbuf, vid);
 	}
 }
 
-static void print_vdi_tree(uint32_t vid, char *name, uint32_t tag,
+static void print_vdi_tree(uint32_t vid, char *name, char * tag, uint32_t snapid,
 			   uint32_t flags, struct sheepdog_inode *i, void *data)
 {
 	time_t ti;
@@ -406,7 +410,7 @@ static void print_vdi_tree(uint32_t vid, char *name, uint32_t tag,
 	add_vdi_tree(name, buf, vid, i->parent_vdi_id, highlight && is_current(i));
 }
 
-static void print_vdi_graph(uint32_t vid, char *name, uint32_t tag,
+static void print_vdi_graph(uint32_t vid, char *name, char * tag, uint32_t snapid,
 			    uint32_t flags, struct sheepdog_inode *i, void *data)
 {
 	time_t ti;
@@ -430,7 +434,7 @@ static void print_vdi_graph(uint32_t vid, char *name, uint32_t tag,
 	       "size: %10s\\n"
 	       "date: %10s\\n"
 	       "time: %10s",
-	       name, tag, size_str, dbuf, tbuf);
+	       name, snapid, size_str, dbuf, tbuf);
 
 	if (is_current(i))
 		printf("\",\n    color=\"red\"\n  ];\n\n");
@@ -439,8 +443,9 @@ static void print_vdi_graph(uint32_t vid, char *name, uint32_t tag,
 
 }
 
-static void cal_total_vdi_size(uint32_t vid, char *name, uint32_t tag,
-			       uint32_t flags, struct sheepdog_inode *i, void *data)
+static void cal_total_vdi_size(uint32_t vid, char *name, char * tag,
+			       uint32_t snapid, uint32_t flags,
+			       struct sheepdog_inode *i, void *data)
 {
 	uint64_t *size = data;
 
@@ -450,18 +455,22 @@ static void cal_total_vdi_size(uint32_t vid, char *name, uint32_t tag,
 
 struct get_vid_info {
 	char *name;
+	char *tag;
 	uint32_t vid;
+	uint32_t snapid;
 };
 
-static void get_oid(uint32_t vid, char *name, uint32_t tag,
+static void get_oid(uint32_t vid, char *name, char *tag, uint32_t snapid,
 		    uint32_t flags, struct sheepdog_inode *i, void *data)
 {
 	struct get_vid_info *info = data;
-	int snapshot_id = vdi_cmd_data.snapshot;
 
 	if (info->name) {
-		if (snapshot_id) {
-			if (!strcmp(name, info->name) && tag == snapshot_id)
+		if (info->tag) {
+			if (!strcmp(name, info->name) && !strcmp(tag, info->tag))
+				info->vid = vid;
+		} else if (info->snapid) {
+			if (!strcmp(name, info->name) && snapid == info->snapid)
 				info->vid = vid;
 		} else {
 			if (!strcmp(name, info->name))
@@ -717,7 +726,7 @@ static int vdi_delete(int argc, char **argv)
 	struct sd_vdi_req hdr;
 	struct sd_vdi_rsp *rsp = (struct sd_vdi_rsp *)&hdr;
 	unsigned rlen, wlen;
-	char vdiname[SD_MAX_VDI_LEN];
+	char vdiname[SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN];
 
 	fd = connect_to(sdhost, sdport);
 	if (fd < 0)
@@ -729,11 +738,14 @@ static int vdi_delete(int argc, char **argv)
 	wlen = sizeof(vdiname);
 
 	hdr.opcode = SD_OP_DEL_VDI;
-	hdr.snapid = vdi_cmd_data.snapshot;
+	hdr.snapid = vdi_cmd_data.snapshot_id;
 	hdr.epoch = node_list_version;
 	hdr.flags = SD_FLAG_CMD_WRITE;
 	hdr.data_length = wlen;
-	strncpy(vdiname, data, sizeof(vdiname));
+	memset(vdiname, 0, sizeof(vdiname));
+	strncpy(vdiname, data, SD_MAX_VDI_LEN);
+	strncpy(vdiname + SD_MAX_VDI_LEN, vdi_cmd_data.snapshot_tag,
+		SD_MAX_VDI_TAG_LEN);
 
 	ret = exec_req(fd, (struct sd_req *)&hdr, vdiname, &wlen, &rlen);
 	close(fd);
@@ -759,8 +771,11 @@ static int vdi_object(int argc, char **argv)
 	struct get_vid_info info;
 	uint32_t vid;
 
+	memset(&info, 0, sizeof(info));
 	info.name = vdiname;
+	info.tag = vdi_cmd_data.snapshot_tag;
 	info.vid = 0;
+	info.snapid = vdi_cmd_data.snapshot_id;
 
 	ret = parse_vdi(get_oid, SD_INODE_HEADER_SIZE, &info);
 
@@ -804,12 +819,255 @@ static int vdi_object(int argc, char **argv)
 	return 0;
 }
 
+static int find_vdi_attr_oid(char *vdiname, char *tag, uint32_t snapid,
+			     char *key, uint32_t *vid, uint64_t *oid,
+			     unsigned int *nr_copies, int creat, int excl)
+{
+	struct sd_vdi_req hdr;
+	struct sd_vdi_rsp *rsp = (struct sd_vdi_rsp *)&hdr;
+	int fd, ret;
+	unsigned int wlen, rlen;
+	char buf[SD_ATTR_HEADER_SIZE];
+
+	memset(buf, 0, sizeof(buf));
+	strncpy(buf, vdiname, SD_MAX_VDI_LEN);
+	strncpy(buf + SD_MAX_VDI_LEN, vdi_cmd_data.snapshot_tag,
+		SD_MAX_VDI_TAG_LEN);
+	memcpy(buf + SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN,
+	       &vdi_cmd_data.snapshot_id, sizeof(uint32_t));
+	strncpy(buf + SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN + sizeof(uint32_t),
+		key, SD_MAX_VDI_ATTR_KEY_LEN);
+
+	fd = connect_to(sdhost, sdport);
+	if (fd < 0) {
+		fprintf(stderr, "failed to connect\n\n");
+		return SD_RES_EIO;
+	}
+
+	memset(&hdr, 0, sizeof(hdr));
+	hdr.opcode = SD_OP_GET_VDI_ATTR;
+	wlen = SD_ATTR_HEADER_SIZE;
+	rlen = 0;
+	hdr.proto_ver = SD_PROTO_VER;
+	hdr.data_length = wlen;
+	hdr.snapid = vdi_cmd_data.snapshot_id;
+	hdr.flags = SD_FLAG_CMD_WRITE;
+	if (creat)
+		hdr.flags |= SD_FLAG_CMD_CREAT;
+	if (excl)
+		hdr.flags |= SD_FLAG_CMD_EXCL;
+
+	ret = exec_req(fd, (struct sd_req *)&hdr, buf, &wlen, &rlen);
+	if (ret) {
+		ret = SD_RES_EIO;
+		goto out;
+	}
+
+	if (rsp->result != SD_RES_SUCCESS) {
+		ret = rsp->result;
+		goto out;
+	}
+
+	*vid = rsp->vdi_id;
+	*oid = vid_to_attr_oid(rsp->vdi_id, rsp->attr_id);
+	*nr_copies = rsp->copies;
+
+	ret = SD_RES_SUCCESS;
+out:
+	close(fd);
+	return ret;
+}
+
+static int vdi_setattr(int argc, char **argv)
+{
+	struct sd_obj_req hdr;
+	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
+	char name[128];
+	int i = 0, n, fd, ret;
+	uint64_t oid, attr_oid = 0;
+	uint32_t vid = 0, nr_copies = 0;
+	char *vdiname = argv[optind++], *key, *value;
+	unsigned int wlen = 0, rlen = 0;
+	uint64_t offset;
+
+	key = argv[optind++];
+	if (!key) {
+		fprintf(stderr, "please specify the name of key\n");
+		return 1;
+	}
+
+	value = argv[optind++];
+	if (!value) {
+		value = malloc(SD_MAX_VDI_ATTR_VALUE_LEN);
+		if (!value) {
+			fprintf(stderr, "failed to allocate memory\n");
+			return 1;
+		}
+
+		offset = 0;
+reread:
+		ret = read(STDIN_FILENO, value + offset,
+			   SD_MAX_VDI_ATTR_VALUE_LEN - offset);
+		if (ret < 0) {
+			fprintf(stderr, "failed to read from stdin, %m\n");
+			return 1;
+		}
+		if (ret > 0) {
+			offset += ret;
+			goto reread;
+		}
+	}
+
+	ret = find_vdi_attr_oid(vdiname, vdi_cmd_data.snapshot_tag,
+				vdi_cmd_data.snapshot_id, key, &vid, &attr_oid,
+				&nr_copies, !vdi_cmd_data.delete,
+				vdi_cmd_data.exclusive);
+	if (ret) {
+		if (ret == SD_RES_VDI_EXIST) {
+			fprintf(stderr, "the attribute already exists, %s\n", key);
+		} else if (ret == SD_RES_NO_OBJ) {
+			fprintf(stderr, "no such attribute, %s\n", key);
+		} else
+			fprintf(stderr, "failed to find attr oid, %s\n",
+				sd_strerror(ret));
+		return 1;
+	}
+
+	oid = attr_oid;
+	for (i = 0; i < nr_copies; i++) {
+		rlen = 0;
+		if (vdi_cmd_data.delete)
+			wlen = 1;
+		else
+			wlen = strlen(value);
+
+		n = obj_to_sheep(node_list_entries, nr_nodes, oid, i);
+
+		addr_to_str(name, sizeof(name), node_list_entries[n].addr, 0);
+
+		fd = connect_to(name, node_list_entries[n].port);
+		if (fd < 0) {
+			printf("%s(%d): %s, %m\n", __func__, __LINE__,
+			       name);
+			break;
+		}
+
+		memset(&hdr, 0, sizeof(hdr));
+		hdr.epoch = node_list_version;
+		hdr.opcode = SD_OP_WRITE_OBJ;
+		hdr.oid = oid;
+
+		hdr.data_length = wlen;
+		if (vdi_cmd_data.delete) {
+			hdr.flags =  SD_FLAG_CMD_DIRECT | SD_FLAG_CMD_WRITE;
+			hdr.offset = offsetof(struct sheepdog_inode, name);
+			value = (char *)"";
+		} else {
+			hdr.flags =  SD_FLAG_CMD_DIRECT | SD_FLAG_CMD_WRITE |
+				SD_FLAG_CMD_TRUNCATE;
+			hdr.offset = SD_ATTR_HEADER_SIZE;
+		}
+
+		ret = exec_req(fd, (struct sd_req *)&hdr, value, &wlen, &rlen);
+		close(fd);
+
+		if (ret) {
+			fprintf(stderr, "failed to set attribute\n");
+			return 1;
+		}
+		if (rsp->result != SD_RES_SUCCESS) {
+			fprintf(stderr, "failed to set attribute, %s\n",
+				sd_strerror(rsp->result));
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
+static int vdi_getattr(int argc, char **argv)
+{
+	struct sd_obj_req hdr;
+	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
+	char name[128];
+	int i = 0, n, fd, ret;
+	uint64_t oid, attr_oid = 0;
+	uint32_t vid = 0, nr_copies = 0;
+	char *vdiname = argv[optind++], *key, *value;
+	unsigned int wlen = 0, rlen = 0;
+
+	key = argv[optind++];
+	if (!key) {
+		fprintf(stderr, "please specify the name of key\n");
+		return 1;
+	}
+
+	ret = find_vdi_attr_oid(vdiname, vdi_cmd_data.snapshot_tag,
+				vdi_cmd_data.snapshot_id, key, &vid, &attr_oid,
+				&nr_copies, 0, 0);
+	if (ret == SD_RES_NO_OBJ) {
+		fprintf(stderr, "no such attribute, %s\n", key);
+		return 1;
+	} else if (ret) {
+		fprintf(stderr, "failed to find attr oid, %s\n",
+			sd_strerror(ret));
+		return 1;
+	}
+
+	oid = attr_oid;
+	value = malloc(SD_MAX_VDI_ATTR_VALUE_LEN);
+	if (!value) {
+		fprintf(stderr, "failed to allocate memory\n");
+		return 1;
+	}
+	for (i = 0; i < nr_copies; i++) {
+		rlen = SD_MAX_VDI_ATTR_VALUE_LEN;
+		wlen = 0;
+
+		n = obj_to_sheep(node_list_entries, nr_nodes, oid, i);
+
+		addr_to_str(name, sizeof(name), node_list_entries[n].addr, 0);
+
+		fd = connect_to(name, node_list_entries[n].port);
+		if (fd < 0) {
+			printf("%s(%d): %s, %m\n", __func__, __LINE__,
+			       name);
+			goto out;
+		}
+
+		memset(&hdr, 0, sizeof(hdr));
+		hdr.epoch = node_list_version;
+		hdr.opcode = SD_OP_READ_OBJ;
+		hdr.oid = oid;
+
+		hdr.data_length = rlen;
+		hdr.flags =  SD_FLAG_CMD_DIRECT;
+		hdr.offset = SD_ATTR_HEADER_SIZE;
+
+		ret = exec_req(fd, (struct sd_req *)&hdr, value, &wlen, &rlen);
+		close(fd);
+
+		if (!ret) {
+			if (rsp->result == SD_RES_SUCCESS) {
+				printf("%s", value);
+				free(value);
+				return 0;
+			}
+		}
+	}
+out:
+	free(value);
+	return 1;
+}
+
 static struct subcommand vdi_cmd[] = {
 	{"delete", SUBCMD_FLAG_NEED_NOEDLIST|SUBCMD_FLAG_NEED_THIRD_ARG, vdi_delete},
 	{"list", SUBCMD_FLAG_NEED_NOEDLIST, vdi_list},
 	{"tree", SUBCMD_FLAG_NEED_NOEDLIST, vdi_tree},
 	{"graph", SUBCMD_FLAG_NEED_NOEDLIST, vdi_graph},
 	{"object", SUBCMD_FLAG_NEED_NOEDLIST|SUBCMD_FLAG_NEED_THIRD_ARG, vdi_object},
+	{"setattr", SUBCMD_FLAG_NEED_NOEDLIST|SUBCMD_FLAG_NEED_THIRD_ARG, vdi_setattr},
+	{"getattr", SUBCMD_FLAG_NEED_NOEDLIST|SUBCMD_FLAG_NEED_THIRD_ARG, vdi_getattr},
 	{NULL,},
 };
 
@@ -818,6 +1076,8 @@ static struct option vdi_long_options[] =
 	COMMON_LONG_OPTIONS
 	{"index", required_argument, NULL, 'i'},
 	{"snapshot", required_argument, NULL, 's'},
+	{"exclusive", no_argument, NULL, 'x'},
+	{"delete", no_argument, NULL, 'd'},
 	{NULL, 0, NULL, 0},
 };
 
@@ -828,7 +1088,16 @@ static int vdi_parser(int ch, char *opt)
 		vdi_cmd_data.index = atoi(opt);
 		break;
 	case 's':
-		vdi_cmd_data.snapshot = atoi(opt);
+		vdi_cmd_data.snapshot_id = atoi(opt);
+		if (vdi_cmd_data.snapshot_id == 0)
+			strncpy(vdi_cmd_data.snapshot_tag, opt,
+				sizeof(vdi_cmd_data.snapshot_tag));
+		break;
+	case 'x':
+		vdi_cmd_data.exclusive = 1;
+		break;
+	case 'd':
+		vdi_cmd_data.delete = 1;
 		break;
 	}
 
@@ -938,7 +1207,7 @@ static struct {
 } commands[] = {
 	{"vdi", vdi_cmd,
 	 vdi_long_options,
-	 COMMON_SHORT_OPTIONS "i:s:",
+	 COMMON_SHORT_OPTIONS "i:s:xd",
 	 vdi_parser,},
 	{"node", node_cmd,},
 	{"cluster", cluster_cmd,
