@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -632,16 +632,17 @@ static void vdi_op(struct vdi_op_message *msg)
 	struct sd_vdi_rsp *rsp = &msg->rsp;
 	void *data = msg->data;
 	int ret = SD_RES_SUCCESS;
-	uint32_t vid = 0;
+	uint32_t vid = 0, attrid = 0, nr_copies = sys->nr_sobjs;
 
 	switch (hdr->opcode) {
 	case SD_OP_NEW_VDI:
 		ret = add_vdi(hdr->epoch, data, hdr->data_length, hdr->vdi_size, &vid,
 			      hdr->base_vdi_id, hdr->copies,
-			      hdr->snapid);
+			      hdr->snapid, &nr_copies);
 		break;
 	case SD_OP_DEL_VDI:
-		ret = del_vdi(hdr->epoch, data, hdr->data_length, &vid, hdr->snapid);
+		ret = del_vdi(hdr->epoch, data, hdr->data_length, &vid,
+			      hdr->snapid, &nr_copies);
 		break;
 	case SD_OP_LOCK_VDI:
 	case SD_OP_GET_VDI_INFO:
@@ -649,10 +650,21 @@ static void vdi_op(struct vdi_op_message *msg)
 			ret = SD_RES_VER_MISMATCH;
 			break;
 		}
-		ret = lookup_vdi(hdr->epoch, data, hdr->data_length, &vid, hdr->snapid);
+		ret = lookup_vdi(hdr->epoch, data, hdr->data_length, &vid,
+				 hdr->snapid, &nr_copies);
 		if (ret != SD_RES_SUCCESS)
 			break;
 		break;
+	case SD_OP_GET_VDI_ATTR:
+		ret = lookup_vdi(hdr->epoch, data,
+				 min(SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN, hdr->data_length),
+				 &vid, hdr->snapid, &nr_copies);
+		if (ret != SD_RES_SUCCESS)
+			break;
+		ret = get_vdi_attr(hdr->epoch, data, hdr->data_length, vid,
+				   &attrid, hdr->flags & SD_FLAG_CMD_CREAT,
+				   hdr->flags & SD_FLAG_CMD_EXCL);
+		break;
 	case SD_OP_RELEASE_VDI:
 		break;
 	case SD_OP_MAKE_FS:
@@ -667,6 +679,8 @@ static void vdi_op(struct vdi_op_message *msg)
 	}
 
 	rsp->vdi_id = vid;
+	rsp->attr_id = attrid;
+	rsp->copies = nr_copies;
 	rsp->result = ret;
 }
 
@@ -697,6 +711,7 @@ static void vdi_op_done(struct vdi_op_message *msg)
 	case SD_OP_LOCK_VDI:
 	case SD_OP_RELEASE_VDI:
 	case SD_OP_GET_VDI_INFO:
+	case SD_OP_GET_VDI_ATTR:
 		break;
 	case SD_OP_MAKE_FS:
 		sys->nr_sobjs = ((struct sd_so_req *)hdr)->copies;
@@ -1233,7 +1248,7 @@ int is_access_to_busy_objects(uint64_t oid)
 				eprintf("bug\n");
 			continue;
 		}
-		if (oid == req->local_oid[0] || oid == req->local_oid[1])
+		if (oid == req->local_oid)
 				return 1;
 	}
 	return 0;
@@ -1247,12 +1262,10 @@ static int __is_access_to_busy_objects(struct request *req)
 		return 0;
 	}
 
-	if (is_access_to_busy_objects(req->local_oid[0]) ||
-	    is_access_to_busy_objects(req->local_oid[1]))
+	if (is_access_to_busy_objects(req->local_oid))
 		return 1;
 
-	if (is_recoverying_oid(req->local_oid[0]) ||
-	    is_recoverying_oid(req->local_oid[1]))
+	if (is_recoverying_oid(req->local_oid))
 		return 1;
 
 	return 0;
