@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -72,11 +72,7 @@ static void setup_access_to_local_objects(struct request *req)
 	int copies;
 
 	if (hdr->flags & SD_FLAG_CMD_DIRECT) {
-		req->local_oid[0] = hdr->oid;
-
-		if (hdr->flags & SD_FLAG_CMD_COW)
-			req->local_oid[1] = hdr->cow_oid;
-
+		req->local_oid = hdr->oid;
 		return;
 	}
 
@@ -85,11 +81,7 @@ static void setup_access_to_local_objects(struct request *req)
 		copies = sys->nr_sobjs;
 
 	if (is_access_local(req->entry, req->nr_nodes, hdr->oid, copies))
-		req->local_oid[0] = hdr->oid;
-
-	if ((hdr->flags & SD_FLAG_CMD_COW) &&
-	    is_access_local(req->entry, req->nr_nodes, hdr->cow_oid, copies))
-		req->local_oid[1] = hdr->cow_oid;
+		req->local_oid = hdr->oid;
 }
 
 static void __done(struct work *work, int idx)
@@ -106,6 +98,7 @@ static void __done(struct work *work, int idx)
 	case SD_OP_GET_VDI_INFO:
 	case SD_OP_MAKE_FS:
 	case SD_OP_SHUTDOWN:
+	case SD_OP_GET_VDI_ATTR:
 		/* request is forwarded to cpg group */
 		return;
 	}
@@ -232,6 +225,7 @@ static void queue_request(struct request *req)
 	case SD_OP_MAKE_FS:
 	case SD_OP_SHUTDOWN:
 	case SD_OP_STAT_CLUSTER:
+	case SD_OP_GET_VDI_ATTR:
 		req->work.fn = cluster_queue_request;
 		break;
 	case SD_OP_READ_VDIS:
@@ -645,7 +639,7 @@ int read_object(struct sheepdog_node_list_entry *e,
 	struct sd_obj_req hdr;
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
 	char name[128];
-	int i = 0, n, fd, ret;
+	int i = 0, n, fd, ret, last_error = SD_RES_SUCCESS;
 
 	if (nr > nodes)
 		nr = nodes;
@@ -664,13 +658,6 @@ int read_object(struct sheepdog_node_list_entry *e,
 				return ret;
 			}
 
-			/* fix replication consistency */
-			if (nr > 1) {
-				if (write_object(e, nodes, node_version, oid,
-						 data, datalen, offset, nr, 0))
-					eprintf("failed to fix obj consistency\n");
-			}
-
 			return ret;
 		}
 
@@ -687,7 +674,7 @@ int read_object(struct sheepdog_node_list_entry *e,
 		if (fd < 0) {
 			printf("%s(%d): %s, %m\n", __func__, __LINE__,
 			       name);
-			return -1;
+			return -SD_RES_EIO;
 		}
 
 		memset(&hdr, 0, sizeof(hdr));
@@ -702,13 +689,18 @@ int read_object(struct sheepdog_node_list_entry *e,
 		ret = exec_req(fd, (struct sd_req *)&hdr, data, &wlen, &rlen);
 		close(fd);
 
-		if (!ret) {
-			if (rsp->result == SD_RES_SUCCESS)
-				return rsp->data_length;
+		if (ret) {
+			last_error = SD_RES_EIO;
+			continue;
 		}
+
+		if (rsp->result == SD_RES_SUCCESS)
+			return rsp->data_length;
+
+		last_error = rsp->result;
 	}
 
-	return -1;
+	return -last_error;
 }
 
 int remove_object(struct sheepdog_node_list_entry *e,
