@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -81,7 +81,7 @@ struct request {
 	struct list_head r_wlist;
 	struct list_head pending_list;
 
-	uint64_t local_oid[2];
+	uint64_t local_oid;
 
 	struct sheepdog_node_list_entry entry[SD_MAX_NODES];
 	int nr_nodes;
@@ -144,14 +144,19 @@ int init_store(const char *dir);
 
 int add_vdi(uint32_t epoch, char *data, int data_len, uint64_t size,
 	    uint32_t *new_vid, uint32_t base_vid, uint32_t copies,
-	    int is_snapshot);
+	    int is_snapshot, unsigned int *nr_copies);
 
-int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t snapid);
+int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid,
+	    uint32_t snapid, unsigned int *nr_copies);
 
-int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t snapid);
+int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid,
+	       uint32_t snapid, unsigned int *nr_copies);
 
 int read_vdis(char *data, int len, unsigned int *rsp_len);
 
+int get_vdi_attr(uint32_t epoch, char *data, int data_len, uint32_t vid,
+		 uint32_t *attrid, int creat, int excl);
+
 int setup_ordered_sd_node_list(struct request *req);
 int get_ordered_sd_node_list(struct sheepdog_node_list_entry *entries);
 int is_access_to_busy_objects(uint64_t oid);
