@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -47,7 +47,7 @@ static int create_vdi_obj(uint32_t epoch, char *name, uint32_t new_vid, uint64_t
 
 			ret = read_object(entries, nr_nodes, epoch,
 					  vid_to_vdi_oid(cur_vid), (char *)&cur,
-					  sizeof(cur), 0, copies);
+					  SD_INODE_HEADER_SIZE, 0, copies);
 			if (ret < 0) {
 				vprintf(SDOG_ERR "failed\n");
 				return SD_RES_BASE_VDI_READ;
@@ -90,7 +90,7 @@ static int create_vdi_obj(uint32_t epoch, char *name, uint32_t new_vid, uint64_t
 	if (is_snapshot && cur_vid != base_vid) {
 		ret = write_object(entries, nr_nodes, epoch,
 				   vid_to_vdi_oid(cur_vid), (char *)&cur,
-				   sizeof(cur), 0, copies, 0);
+				   SD_INODE_HEADER_SIZE, 0, copies, 0);
 		if (ret < 0) {
 			vprintf(SDOG_ERR "failed\n");
 			return SD_RES_BASE_VDI_READ;
@@ -100,7 +100,7 @@ static int create_vdi_obj(uint32_t epoch, char *name, uint32_t new_vid, uint64_t
 	if (base_vid) {
 		ret = write_object(entries, nr_nodes, epoch,
 				   vid_to_vdi_oid(base_vid), (char *)&base,
-				   sizeof(base), 0, copies, 0);
+				   SD_INODE_HEADER_SIZE, 0, copies, 0);
 		if (ret < 0) {
 			vprintf(SDOG_ERR "failed\n");
 			return SD_RES_BASE_VDI_WRITE;
@@ -118,7 +118,8 @@ static int create_vdi_obj(uint32_t epoch, char *name, uint32_t new_vid, uint64_t
 
 static int find_first_vdi(uint32_t epoch, unsigned long start, unsigned long end,
 			  char *name, char *tag, uint32_t snapid, uint32_t *vid,
-			  unsigned long *deleted_nr, uint32_t *next_snap)
+			  unsigned long *deleted_nr, uint32_t *next_snap,
+			  unsigned int *nr_copies)
 {
 	struct sheepdog_node_list_entry entries[SD_MAX_NODES];
 	static struct sheepdog_inode inode;
@@ -135,7 +136,7 @@ static int find_first_vdi(uint32_t epoch, unsigned long start, unsigned long end
 	for (i = start; i >= end; i--) {
 		ret = read_object(entries, nr_nodes, epoch,
 				  vid_to_vdi_oid(i), (char *)&inode,
-				  sizeof(inode), 0, nr_reqs);
+				  SD_INODE_HEADER_SIZE, 0, nr_reqs);
 		if (ret < 0)
 			return SD_RES_EIO;
 
@@ -154,6 +155,7 @@ static int find_first_vdi(uint32_t epoch, unsigned long start, unsigned long end
 
 			*next_snap = inode.snap_id + 1;
 			*vid = inode.vdi_id;
+			*nr_copies = inode.nr_copies;
 			return SD_RES_SUCCESS;
 		}
 	}
@@ -167,7 +169,8 @@ static int find_first_vdi(uint32_t epoch, unsigned long start, unsigned long end
 
 static int do_lookup_vdi(uint32_t epoch, char *name, int namelen, uint32_t *vid,
 			 char *tag, uint32_t snapid, uint32_t *next_snapid,
-			 unsigned long *right_nr,  unsigned long *deleted_nr)
+			 unsigned long *right_nr,  unsigned long *deleted_nr,
+			 unsigned int *nr_copies)
 {
 	int ret;
 	unsigned long nr, start_nr;
@@ -185,7 +188,7 @@ static int do_lookup_vdi(uint32_t epoch, char *name, int namelen, uint32_t *vid,
 	right_side:
 		/* look up on the right side of the hash point */
 		ret = find_first_vdi(epoch, nr - 1, start_nr, name, tag, snapid, vid,
-				     deleted_nr, next_snapid);
+				     deleted_nr, next_snapid, nr_copies);
 		return ret;
 	} else {
 		/* round up... bitmap search from the head of the bitmap */
@@ -196,7 +199,7 @@ static int do_lookup_vdi(uint32_t epoch, char *name, int namelen, uint32_t *vid,
 		else if (nr) {
 			/* look up on the left side of the hash point */
 			ret = find_first_vdi(epoch, nr - 1, 0, name, tag, snapid, vid,
-					     deleted_nr, next_snapid);
+					     deleted_nr, next_snapid, nr_copies);
 			if (ret == SD_RES_NO_VDI)
 				; /* we need to go to the right side */
 			else
@@ -208,7 +211,8 @@ static int do_lookup_vdi(uint32_t epoch, char *name, int namelen, uint32_t *vid,
 	}
 }
 
-int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t snapid)
+int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid,
+	       uint32_t snapid, unsigned int *nr_copies)
 {
 	char *name = data, *tag;
 	uint32_t dummy0;
@@ -222,11 +226,12 @@ int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t
 		return SD_RES_INVALID_PARMS;
 
 	return do_lookup_vdi(epoch, name, strlen(name), vid, tag, snapid,
-			     &dummy0, &dummy1, &dummy2);
+			     &dummy0, &dummy1, &dummy2, nr_copies);
 }
 
 int add_vdi(uint32_t epoch, char *data, int data_len, uint64_t size,
-	    uint32_t *new_vid, uint32_t base_vid, uint32_t copies, int is_snapshot)
+	    uint32_t *new_vid, uint32_t base_vid, uint32_t copies,
+	    int is_snapshot, unsigned int *nr_copies)
 {
 	uint32_t cur_vid;
 	uint32_t next_snapid;
@@ -240,7 +245,7 @@ int add_vdi(uint32_t epoch, char *data, int data_len, uint64_t size,
 	name = data;
 
 	ret = do_lookup_vdi(epoch, name, strlen(name), &cur_vid, NULL, 0, &next_snapid,
-			    &right_nr, &deleted_nr);
+			    &right_nr, &deleted_nr, nr_copies);
 
 	if (is_snapshot) {
 		if (ret != SD_RES_SUCCESS) {
@@ -285,9 +290,10 @@ int add_vdi(uint32_t epoch, char *data, int data_len, uint64_t size,
 
 int start_deletion(uint32_t vid, uint32_t epoch);
 
-int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t snapid)
+int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid,
+	    uint32_t snapid, unsigned int *nr_copies)
 {
-	char *name = data;
+	char *name = data, *tag;
 	uint32_t dummy0;
 	unsigned long dummy1, dummy2;
 	int ret;
@@ -295,11 +301,15 @@ int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t sn
 	int nr_nodes, nr_reqs;
 	static struct sheepdog_inode inode;
 
-	if (data_len != SD_MAX_VDI_LEN)
+	if (data_len == SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN)
+		tag = data + SD_MAX_VDI_LEN;
+	else if (data_len == SD_MAX_VDI_LEN)
+		tag = NULL;
+	else
 		return SD_RES_INVALID_PARMS;
 
-	ret = do_lookup_vdi(epoch, name, strlen(name), vid, NULL, snapid,
-			     &dummy0, &dummy1, &dummy2);
+	ret = do_lookup_vdi(epoch, name, strlen(name), vid, tag, snapid,
+			    &dummy0, &dummy1, &dummy2, nr_copies);
 	if (ret != SD_RES_SUCCESS)
 		return ret;
 
@@ -309,22 +319,22 @@ int del_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid, uint32_t sn
 		nr_reqs = nr_nodes;
 
 	ret = read_object(entries, nr_nodes, epoch,
-			  vid_to_vdi_oid(*vid), (char *)&inode, sizeof(inode), 0,
-			  nr_reqs);
+			  vid_to_vdi_oid(*vid), (char *)&inode,
+			  SD_INODE_HEADER_SIZE, 0, nr_reqs);
 	if (ret < 0)
 		return SD_RES_EIO;
 
 	memset(inode.name, 0, sizeof(inode.name));
 
 	ret = write_object(entries, nr_nodes, epoch,
-			   vid_to_vdi_oid(*vid), (char *)&inode, sizeof(inode), 0,
-			   nr_reqs, 0);
+			   vid_to_vdi_oid(*vid), (char *)&inode,
+			   SD_INODE_HEADER_SIZE, 0, nr_reqs, 0);
 	if (ret < 0)
 		return SD_RES_EIO;
 
 	ret = start_deletion(*vid, epoch);
-	if (ret < 0)
-		return SD_RES_NO_MEM;
+	if (ret != SD_RES_SUCCESS)
+		return ret;
 
 	return SD_RES_SUCCESS;
 }
@@ -428,10 +438,10 @@ static int fill_vdi_list(struct deletion_work *dw,
 again:
 	vid = ((uint32_t *)dw->buf)[done++];
 	ret = read_object(entries, nr_entries, dw->epoch,
-			  vid_to_vdi_oid(vid), (void *)&inode, sizeof(inode),
-			  0, nr_entries);
+			  vid_to_vdi_oid(vid), (void *)&inode,
+			  SD_INODE_HEADER_SIZE, 0, sys->nr_sobjs);
 
-	if (ret != sizeof(inode)) {
+	if (ret != SD_INODE_HEADER_SIZE) {
 		eprintf("cannot find vdi object\n");
 		return 0;
 	}
@@ -460,10 +470,10 @@ static uint64_t get_vdi_root(struct sheepdog_node_list_entry *entries,
 
 next:
 	ret = read_object(entries, nr_entries, epoch,
-			  vid_to_vdi_oid(vid),
-			  (void *)&inode, sizeof(inode), 0, nr_entries);
+			  vid_to_vdi_oid(vid), (void *)&inode,
+			  SD_INODE_HEADER_SIZE, 0, sys->nr_sobjs);
 
-	if (ret != sizeof(inode)) {
+	if (ret != SD_INODE_HEADER_SIZE) {
 		eprintf("cannot find vdi object\n");
 		return 0;
 	}
@@ -485,12 +495,12 @@ int start_deletion(uint32_t vid, uint32_t epoch)
 
 	dw = zalloc(sizeof(struct deletion_work));
 	if (!dw)
-		return -1;
+		return SD_RES_NO_MEM;
 
 	dw->buf = zalloc(1 << 20); /* FIXME: handle larger buffer */
 	if (!dw->buf) {
 		free(dw);
-		return -1;
+		return SD_RES_NO_MEM;
 	}
 
 	dw->count = 0;
@@ -504,24 +514,82 @@ int start_deletion(uint32_t vid, uint32_t epoch)
 
 	root_vid = get_vdi_root(entries, nr_nodes, dw->epoch, dw->vid);
 	if (!root_vid)
-		return -1;
+		return SD_RES_EIO;
 
 	ret = fill_vdi_list(dw, entries, nr_nodes, root_vid);
 	if (ret)
-		return 0;
+		return SD_RES_SUCCESS;
 
 	dprintf("%d\n", dw->count);
 
 	if (dw->count == 0)
-		return 0;
+		return SD_RES_SUCCESS;
 
 	if (!list_empty(&deletion_work_list)) {
 		list_add_tail(&dw->dw_siblings, &deletion_work_list);
-		return 0;
+		return SD_RES_SUCCESS;
 	}
 
 	list_add_tail(&dw->dw_siblings, &deletion_work_list);
 	queue_work(&dw->work);
 
-	return 0;
+	return SD_RES_SUCCESS;
+}
+
+int get_vdi_attr(uint32_t epoch, char *data, int data_len, uint32_t vid,
+		 uint32_t *attrid, int creat, int excl)
+{
+	struct sheepdog_node_list_entry entries[SD_MAX_NODES];
+	char attr_buf[SD_ATTR_HEADER_SIZE], inode_buf[SD_INODE_HEADER_SIZE];
+	uint64_t oid;
+	uint32_t end;
+	int ret, nr_nodes, copies;
+
+	if (data_len != SD_ATTR_HEADER_SIZE)
+		return SD_RES_INVALID_PARMS;
+
+	nr_nodes = get_ordered_sd_node_list(entries);
+
+	ret = read_object(entries, nr_nodes, epoch, vid_to_vdi_oid(vid),
+			  inode_buf, sizeof(inode_buf), 0, sys->nr_sobjs);
+	if (ret != SD_INODE_HEADER_SIZE) {
+		eprintf("failed to read vdi object, %"PRIx32"\n", vid);
+		return -ret;
+	}
+
+	copies = ((struct sheepdog_inode *)inode_buf)->nr_copies;
+
+	*attrid = fnv_64a_buf(data, data_len, FNV1A_64_INIT);
+	*attrid &= (UINT64_C(1) << VDI_SPACE_SHIFT) - 1;
+
+	end = *attrid - 1;
+	while (*attrid != end) {
+		oid = vid_to_attr_oid(vid, *attrid);
+		ret = read_object(entries, nr_nodes, epoch, oid, attr_buf,
+				  sizeof(attr_buf), 0, copies);
+
+		if (ret == -SD_RES_NO_OBJ && creat) {
+			ret = write_object(entries, nr_nodes, epoch, oid, data,
+					   data_len, 0, copies, 1);
+			if (ret)
+				return SD_RES_EIO;
+
+			return SD_RES_SUCCESS;
+		}
+
+		if (ret < 0)
+			return -ret;
+
+		if (memcmp(attr_buf, data, sizeof(attr_buf)) == 0) {
+			if (excl)
+				return SD_RES_VDI_EXIST;
+			else
+				return SD_RES_SUCCESS;
+		}
+
+		(*attrid)++;
+	}
+
+	dprintf("there is no space for new vdis\n");
+	return SD_RES_FULL_VDI;
 }
